/*
 * Copyright 1994 Sun Microsystems, Inc. All Rights Reserved
 */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*                   PACKAGE INCLUDE                                           *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
#include "osi_lib.h"
#include "ap_lib.h"
#include "ap_poll.h"

#include <stdio.h>
#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <stropts.h>
#include <string.h>
#include <sys/time.h>
#include <nlist.h>
#include <malloc.h>


/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*                   CONSTANT DEFINITION                                       *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
#define NUM_SELECTORS           4       /* Number of selectors in P-addr */
#define P_SELECTOR_OFFSET       0
#define S_SELECTOR_OFFSET       sizeof(ap_octet_string_t)
#define T_SELECTOR_OFFSET       S_SELECTOR_OFFSET + sizeof(ap_octet_string_t)
#define N_SELECTOR_OFFSET       T_SELECTOR_OFFSET + sizeof(ap_octet_string_t)
#define SELECTORS_OFFSET        N_SELECTOR_OFFSET + sizeof(ap_octet_string_t)

 
/*************************************************************************/
/**     ENCODING DECODING RELATED CONSTANTS                             **/
/*************************************************************************/

/* Supported transfer syntax */
#define ISO_8825        500     /* Basic Encoding of single ASN.1 type */

/* Demo program used abstract syntax */
#define TEST_ACSE_PCI   1000    /* ACSE PCI  { 2 2 1 0 1 } */
#define TEST_ABS_PCI1   1001    /* TEST PCI  { 2 2 1 0 2 } */
#define TEST_ABS_PCI2   1002    /* TEST PCI  { 2 2 1 0 3 } */

/* Demo program used Application Context Name */
#define TEST_APPL_CNTX  2000    /* APPL CNTX { 2 2 1 0 4 } */

/* Demo program used Presentation Context ID */
#define TEST_PCID1      101
#define TEST_PCID2      103
#define TEST_PCID3      105



/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*                   MACRO DEFINITION                                          *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
#define tm_malloc(S, P)                                           \
{                                                                 \
    if ( ((P) = (char *) malloc((S))) != NULL )                   \
        memset ( (P), 0, (S) );                                   \
    else printf ("Memory allocation failed.");                \
} /* tm_malloc */



#define tm_free(P)                                                \
{                                                                 \
    if ( (P) != NULL ) {                                          \
        free( (char *)(P) );                                      \
        (P) = NULL;                                               \
    }                                                             \
} /* tm_free */



 
/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        set_cntx_name()                                       *
*                                                                             *
* Description                                                                 *
*       Encode an application context name in object identifier form          *
*       by calling encode_objid(), then invoke ap_set_env() on env.           *
*       AP_CNTX_NAME.                                                         *
*                                                                             *
* Input Parameters                                                            *
*       None                                                                  *
*                                                                             *
* Output Parameters                                                           *
*       None                                                                  *
*                                                                             *
* Called By                                                                   *
*       main (initiator)                                                      *
*                                                                             *
* External Calls                                                              *
*       encode_objid()  - An object id encoding routine in asn1_if.c.         *
*       ap_set_env()    - An ACSE/Presentation external function to set an    *
*                         environment attribute.                              *
*                                                                             *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
set_cntx_name(fd)
int             fd;
{
    int         rc;             /* Return code from ap_set_env */
    objid_t     ap_cntx_name;   /* A/P structure to be filled  */
    ap_val_t    val;            /* Value passed to ap_set_env()*/
 
    /********************************************************************/
    /* Encode a proposed application context name                       */
    /********************************************************************/
    if ( encode_objid ( TEST_APPL_CNTX, &ap_cntx_name ) == FAIL )
        printf("Initiator: Build_context failed.");
 
 
    /********************************************************************/
    /* Pass the encoded application context name to ap_set_env() and    */
    /* call free_objid() to free the long buf, if it exists.            */
    /********************************************************************/
    val.v = (void *) &ap_cntx_name;
    rc = ap_set_env (fd, AP_CNTX_NAME, val);
    free_objid ( &ap_cntx_name );
       
 
    /********************************************************************/
    /* Error in ap_set_env(), report an error and return FAIL.          */
    /********************************************************************/
    if ( rc != SUCCESS ) {
        printf ("Initiator: ap_set_env AP_CNTX_NAME failed.");
    }
 
    return ( SUCCESS );
 
} /* set_cntx_name */
 
/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        make_pcdrl()
*
* Description
*       
*       The responder needs to make up a presentation context definition 
*       result list to indicate which transfer syntax associated with an
*       abstract syntax is supported, based on the proposed presentation
*       context definition list from the remote side. This routine will 
*       call ap_get_env() to see what's been proposed then create a
*       result list.    
*
* Input Parameters
*       None
*
* Output Parameters
*       None
*
* Called By
*       main (responder) 
*
* External Calls 
*       ap_get_env()    - an external ACSE/Presentation library function
*                         to retrieve the value of an environment attribute.
*       ap_free()       - an external ACSE/Presentation library function
*                         to free memory allocated by ap_get_env().
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
make_pcdrl(fd)   
int     fd;  
{   
    int         rc;             /* Return code from ap_set_env */
    ap_val_t    val;            /* Value passed to ap_set_env  */
    ap_cdl_t    pcdl;           /* Proposed presentation context definition list */
    ap_cdrl_t   pcdrl;          /* Presentation context definition result list */
                                /* to be filled.                         */

    /********************************************************************/
    /* Get the proposed PCDL from the environment                       */
    /********************************************************************/
    if ( ap_get_env ( fd, AP_PCDL, &pcdl ) != SUCCESS ) {
        printf ("ap_get_env AP_PCDL failed.");
    }
 
 
    /********************************************************************/
    /* Build the PCDRL based on the PCDL                                */
    /********************************************************************/
    rc = build_pcdrl ( &pcdl, &pcdrl );
 
 
    /********************************************************************/
    /* Free the AP_PCDL env. acquired through ap_get_env()              */
    /********************************************************************/
    ap_free ( AP_PCDL, &pcdl );
 
 
    /********************************************************************/
    /* Report an error if building the PCDRL failed                     */
    /********************************************************************/
    if ( rc == FAIL )
        printf ("Failed to build PCDRL.");
 
 
    /********************************************************************/
    /* Set the AP_PCDRL environment attribute                           */
    /********************************************************************/
    val.v = (void *) &pcdrl;
    rc = ap_set_env ( fd, AP_PCDRL, val );

    /********************************************************************/
    /* Set the AP_PCDRL environment attribute                           */
    /********************************************************************/
    val.v = (void *) &pcdrl;
    rc = ap_set_env ( fd, AP_PCDRL, val );
 
 
    /********************************************************************/
    /* Need to use our own free routine to free the  memory allocated   */
    /* in build_pcdrl(), we can't depend on ap_free(), because the way  */
    /* that we allocate memory may not be the same as the way the A/P   */
    /* library allocates its memory.                                    */
    /********************************************************************/
    free_pcdrl ( &pcdrl );


    /********************************************************************/
    /* If error in setting AP_PCDRL, report and return                  */
    /********************************************************************/
    if ( rc == FAIL ) {
        printf ("ap_set_env AP_PCDRL failed.");
    }
    return ( SUCCESS );
 
} /* make_pcdrl */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        build_pcdrl()
*
* Description
*       Based on presentation context definition list to make up a
*       presentation context definition result list.
*
* Input Parameters
*       pcdl    - pointer to presentation context definition list.
*       pcdrl   - pointer to presentation context definition result list
*                 to be built.
*
* Output Parameters
*       pcdrl   - pointer to completed pcdrl.
*
*
* Called By
*       make_pcdrl()
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
build_pcdrl ( pcdl, pcdrl )
ap_cdl_t        *pcdl;
ap_cdrl_t       *pcdrl;
{
    int         size, i, j, found;
    char        *mptr;
    objid_t     *next;
    objid_t     sup_xfr_syntax;
 
    /*******************************************/
    /* Initialize the outer structure of PCDRL */
    /*******************************************/
    memset ( pcdrl, 0, sizeof (ap_cdrl_t) );
 
    /**************************************/
    /* Create a supported transfer syntax */
    /**************************************/
    if ( encode_objid ( ISO_8825, &sup_xfr_syntax ) )
        printf ( "build_pcdrl: Error in encoding supported transfer syntax" );
 
 
    /*****************************************/
    /* Allocate maximum size for PCDRL based */
    /* on the number of elements of PCDL.    */
    /*****************************************/
    size =
        ((sizeof (struct seqof_ap_cdrl) + sizeof(objid_t)) * pcdl->size_ap_cdl);
    tm_malloc ( size, mptr );
       
 
    /**************************************/
    /* Set the starting address for PCDRL */
    /**************************************/
    pcdrl->m_ap_cdrl = (struct seqof_ap_cdrl *) mptr;
 
 
    /************************************/
    /* Reserve space for accepted objid */
    /************************************/
    next = (objid_t *) (mptr + (pcdl->size_ap_cdl * sizeof(struct seqof_ap_cdrl)));
 
 
    /***********************************************************/
    /* Go through all elements in PCDL to compare the transfer */
    /* syntax that we supported and set the result of each     */
    /* element accordingly.                                   */
    /***********************************************************/
    pcdrl->size_ap_cdrl = pcdl->size_ap_cdl;
    for ( i = 0; i < pcdrl->size_ap_cdrl; i++ )
    {
        found = 0;
        for ( j = 0; j < pcdl->m_ap_cdl[i].size_t_sytx; j++ )
        {
            if ( objcmp(pcdl->m_ap_cdl[i].m_t_sytx[j], &sup_xfr_syntax) ) {
                pcdrl->m_ap_cdrl[i].res = ACCPT;
                pcdrl->m_ap_cdrl[i].prov_rsn = RSN_NSPEC;
                pcdrl->m_ap_cdrl[i].t_sytx = next++;
                memcpy (pcdrl->m_ap_cdrl[i].t_sytx, &sup_xfr_syntax,
                    sizeof(objid_t));
                found = 1;
                break;
            }
        }
    }
} /* build_pcdrl */
 
/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        free_pcdrl()
*
* Description
*       Free the internal memory of struct ap_cdrl_t allocated by
*       build_pcdrl().  This memory is allocated in one big piece.
*
* Input Parameters
*       PCDRL   - pointer to memory to be freed.
*
* Output Parameters
*       None
*
*
* Called By
*       make_pcdrl()
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
free_pcdrl (pcdrl)
ap_cdrl_t       *pcdrl;
{
    if ( pcdrl ) {
        tm_free ( pcdrl->m_ap_cdrl );
        memset ( pcdrl, 0, sizeof (ap_cdrl_t) );
    }
} /* Free PCDRL */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        set_pcdl()
*
* Description
*       Set up 3 elements of presentation context definition list by
*       calling build_pcdl() to build the data in the ap_cdl_t structure, 
*       an ACSE/Presentation (A/P) Interface data structure, then invoke
*       ap_set_env() to set the AP_PCDL environment attribute.
*
* Input Parameters
*       None
*
* Output Parameters
*       None
*
*
* Called By
*       main (initiator)
*
* External Calls
*       ap_set_env()    - an ACSE/Presentation external function to set an
*                         environment attribute.
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
set_pcdl(fd)
int     fd;
{
    int         rc;             /* Return code from ap_set_env */
    ap_cdl_t    pcdl;           /* A/P structure to be filled */
    ap_val_t    val;            /* Value passed to ap_set_env  */

    /****************************/
    /* Build 3 elements of PCDL */
    /****************************/
    if ( build_pcdl ( &pcdl ) == FAIL )
        printf ( "Error in build_pcdl" );


    /*******************************************************************/
    /* Pass the presentation context definition list to ap_set_env()   */
    /* and call free_pcdl() to free the inner memory allocated in      */
    /* build_pcdl().                                                   */
    /*******************************************************************/
    val.v = (void *) &pcdl;
    rc = ap_set_env ( fd, AP_PCDL, val );
    free_pcdl ( &pcdl );


    /********************************************************************/
    /* Error in ap_set_env(), report an error and return FAIL.          */
    /********************************************************************/
    if ( rc == FAIL ) {
        printf ( "ap_set_env AP_PCDL failed." );
    }

    return ( SUCCESS );

} /* set_pcdl */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        build_pcdl()
*
* Description
*       Create 3 elements of presentation context definition list.
*
* Input Parameters
*       pcdl            - Pointer to ap_cdl_t to be built.
*
* Output Parameters
*       pcdl            - Pointer to completed ap_cdl_t.
*
*
* Called By
*       set_pcdl()
*
* External Calls
*       encode_objid    - Encoding routine to encode objid type.
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
build_pcdl ( pcdl )
ap_cdl_t        *pcdl;
{
    char        *mptr;  /* Pointer to memory allocated by this routine */
    int         size;   /* Size of memory needed */
    int         index;  /* Index to cdl element array */
    objid_t     *next;  /* Pointer to next available space for objid */
    objid_t     **tsptr;/* Pointer to transfer syntax array */
 
    pcdl->size_ap_cdl = 3;
 
    /*****************************************************************/
    /* Total size = number of pcdl elements multiplied by the sum of */
    /*          structure seqof_ap_cdl, 2 objid_t structures and a   */
    /*          pointer to the objid_t structure.                    */  
    /*****************************************************************/
    size = pcdl->size_ap_cdl * (sizeof(struct seqof_ap_cdl) +
                (sizeof(objid_t) * 2) + sizeof(objid_t *));
                 
    tm_malloc (size, mptr);
    pcdl->m_ap_cdl = (struct seqof_ap_cdl *) mptr;
 
 
    /**************************************************/
    /* Reserve space for objid pointers, ie. transfer */
    /* syntax array and abstract syntaxes.            */ 
    /**************************************************/
    tsptr = (objid_t **) (mptr +
        (pcdl->size_ap_cdl * sizeof(struct seqof_ap_cdl)));
    next = (objid_t *) ((char *)(tsptr + pcdl->size_ap_cdl));
 
 
    /***************************/
    /* Building the 1st record */
    /***************************/
    index = 0;
    pcdl->m_ap_cdl[index].pci = TEST_PCID1;
    pcdl->m_ap_cdl[index].size_t_sytx = 1;
    pcdl->m_ap_cdl[index].a_sytx = next++;
    pcdl->m_ap_cdl[index].m_t_sytx = tsptr;
    *tsptr++ = next++;


 
    /******************************************************/
    /* Encode abstract syntax TEST_ACSE_PCI { 2 2 1 0 1 } */
    /******************************************************/
    if ( encode_objid (TEST_ACSE_PCI, pcdl->m_ap_cdl[index].a_sytx) == FAIL ) {
        tm_free ( pcdl->m_ap_cdl );
        printf ( "Build_pcdl: Error in encoding the 1st abstract syntax.");
    }
       
 
    /**************************************/
    /* Transfer syntax ISO_8825 ( 2 1 1 ) */
    /**************************************/
    if ( encode_objid (ISO_8825, *pcdl->m_ap_cdl[index].m_t_sytx) == FAIL ) {
        tm_free ( pcdl->m_ap_cdl );
        printf ( "Build_pcdl: Error in encoding the 1st transfer syntax.");
    }
       
 
    /***************************/
    /* Building the 2nd record */
    /***************************/
    index = 1;
    pcdl->m_ap_cdl[index].pci = TEST_PCID2;
    pcdl->m_ap_cdl[index].size_t_sytx = 1;
    pcdl->m_ap_cdl[index].a_sytx = next++;
    pcdl->m_ap_cdl[index].m_t_sytx = tsptr;
    *tsptr++ = next++;
 
 
    /***********************************************/
    /* Abstract syntax { 2 2 1 0 2 } TEST_ABS_PCI2 */
    /***********************************************/
    if ( encode_objid (TEST_ABS_PCI1, pcdl->m_ap_cdl[index].a_sytx) == FAIL ) {
        tm_free ( pcdl->m_ap_cdl );
        printf ( "Build_pcdl: Error in encoding the 2nd abstract syntax.");
    }
       
 
    /**************************************/
    /* Transfer syntax { 2 1 1 } ISO_8825 */
    /**************************************/
    if ( encode_objid ( ISO_8825, *pcdl->m_ap_cdl[index].m_t_sytx) == FAIL ) {
        tm_free ( pcdl->m_ap_cdl );
        printf ( "Build_pcdl: Error in encoding the 2nd transfer syntax.");
    }
       
       
 
    /***************************/
    /* Building the 3rd record */
    /***************************/
    index = 2;
    pcdl->m_ap_cdl[index].pci = TEST_PCID3;
    pcdl->m_ap_cdl[index].size_t_sytx = 1;
    pcdl->m_ap_cdl[index].a_sytx = next++;
    pcdl->m_ap_cdl[index].m_t_sytx = tsptr;
    *tsptr++ = next++;
 
 
    /***********************************************/
    /* Abstract syntax { 2 2 1 0 3 } TEST_ABS_PCI2 */
    /***********************************************/
    if ( encode_objid (TEST_ABS_PCI2, pcdl->m_ap_cdl[index].a_sytx) == FAIL ) {
        tm_free ( pcdl->m_ap_cdl );
        printf ( "Build_pcdl: Error in encoding the 3rd abstract syntax.");
    }
       
 
    /**************************************/
    /* Transfer syntax { 2 1 1 } ISO_8825 */
    /**************************************/
    if ( encode_objid (ISO_8825, *pcdl->m_ap_cdl[index].m_t_sytx) == FAIL ) {
        tm_free ( pcdl->m_ap_cdl );
        printf ( "Build_pcdl: Error in encoding the 3rd transfer syntax.");
    }
       
    return ( SUCCESS );
 
} /* build_pcdl */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        free_pcdl()
*
* Description
*       Free internal memory in structure ap_cdl_t allocated by
*       build_pcdl().  Memory was allocated in one big piece,
*       except the long objid buffer.
*
* Input Parameters
*       pcdl    -       Pointer to memory to be freed.
*
* Output Parameters
*       None
*
*
* Called By
*       set_pcdl()
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
free_pcdl (pcdl)
ap_cdl_t        *pcdl;
{
    int         i,j;
    objid_t     *objid;
 
    if ( pcdl ) {
        if ( pcdl->m_ap_cdl != NULL ) {
 
            /*************************************************************/
            /* Check each element in PCDL to see whether we're using any */
            /* long_buf for extra long object id.  If we do, then free   */
            /* those spaces individually.                                */
            /*************************************************************/
            for ( i = 0; i < pcdl->size_ap_cdl; i++ ) {
        
                /***************************************************/
                /* Free the long buffer in abstract syntax, if any */
                /***************************************************/
                objid = pcdl->m_ap_cdl[i].a_sytx;
                if ( objid->length > MAXOBJBUF )
                    tm_free ( objid->b.long_buf );
 
                /*******************************/
                /* Check all transfer syntaxes */
                /*******************************/
                for ( j = 0; j < pcdl->m_ap_cdl[i].size_t_sytx; j++ )
                {
                    /********************************/
                    /* Free the long buffer, if any */
                    /********************************/
                    objid = pcdl->m_ap_cdl[i].m_t_sytx[j];
                    if ( objid->length > MAXOBJBUF )
                        tm_free ( objid->b.long_buf );
                }
            }
 
            /**************************************/
            /* All elements have been checked, we */
            /* can free the whole thing now.      */
            /**************************************/
            tm_free ( pcdl->m_ap_cdl );
        }
        memset ( pcdl, 0, sizeof ( ap_cdl_t ) );
    }
} /* free_pcdl */
/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        objcmp()
*
* Description
*       Compare two object identifier.
*
* Input Parameters
*       source          - original objid.
*       target          - objid to compare with.
*
* Output Parameters
*       None
*
*
* Called By
*       build_pcdrl(), check_cntx_name().
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
objcmp ( source, target )
objid_t *source, *target;
{
    int size;

    if ( (size = source->length) == target->length ) {
        if ( size > MAXOBJBUF ) /* Compare the long buf */
        {
            if ( memcmp ((char *)source->b.long_buf,
                         (char *)target->b.long_buf,
                         size) == 0 )
                return ( TRUE );
        }
        else
        {
            if ( memcmp ((char *)source->b.short_buf,
                         (char *)target->b.short_buf,
                         size) == 0 )
                return ( TRUE );
        }
    }
    return ( FALSE );
 
} /* objcmp */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        encode_objid()
*
* Description
*       Manually encode object identifier.
*
* Input Parameters
*       kind            - kind of object id.
*       objid           - Pointer to ACSE/Presentation Interface's
*                         object id structure (objid_t).
*
* Output Parameters
*       objid           - Pointer to objid_t contains encoded value.
*
* Globals Referenced
*       msg_buf
*
* Called By (optional)
*       set_cntx_name(), check_cntx_name(), build_pcdl(), build_pcdrl().
*
* External Calls
*       None
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
encode_objid ( kind, objid )
int     kind;
objid_t *objid;
{
    switch ( kind ) {

    case TEST_ACSE_PCI:                 /* abstract syntax */
    case TEST_ABS_PCI1:                 /* abstract syntax */
    case TEST_ABS_PCI2:                 /* abstract syntax */
    case TEST_APPL_CNTX:                /* application context name */
        objid->length = 6;
        objid->b.short_buf[0] = 0x06;
        objid->b.short_buf[1] = 0x04;
        objid->b.short_buf[2] = 0x52;
        objid->b.short_buf[3] = 0x01;
        objid->b.short_buf[4] = 0x00;

        if ( kind == TEST_ACSE_PCI )    /* (ACSE_PCI) 2 2 1 0 1 */
           objid->b.short_buf[5] = 0x01;
        else if ( kind == TEST_ABS_PCI1 )          /* 2 2 1 0 2 */
           objid->b.short_buf[5] = 0x02;
        else if ( kind == TEST_ABS_PCI2 )          /* 2 2 1 0 3 */
           objid->b.short_buf[5] = 0x03;
        else                /* kind = TEST_APPL_CNTX: 2 2 1 0 4 */
           objid->b.short_buf[5] = 0x04;
        break;

    case ISO_8825:                      /* transfer syntax */
        objid->length = 4;
        objid->b.short_buf[0] = 0x06;
        objid->b.short_buf[1] = 0x02;
        objid->b.short_buf[2] = 0x51;
        objid->b.short_buf[3] = 0x01;
        break;

    default:
        printf ("Encode_objid: Unknown object id kind ");
    }
    return (SUCCESS);
 
} /* encode objid */


/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        free_objid()
*
* Description
*       Reset objid to 0 and free the long buffer in objid_t, if it exists.
*
* Input Parameters
*       objid           - Pointer to memory to be freed.
*
* Output Parameters
*       None
*
*
* Called By
*       set_cntx_name())
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
free_objid ( objid )
objid_t *objid;
{
    if ( objid ) {
        if ( objid->length > MAXOBJBUF )
            tm_free ( objid->b.long_buf );
        memset ( objid, 0, sizeof ( objid_t ) );
    }   
} /* free_objid */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* Function Name:        set_paddr()
*
* Description
*       Set a local or remote p-address environment attribute.
*
* Input Parameters
*       attribute       - AP_BIND_PADDR or AP_REM_PADDR.
*       p,s,t,nsap      - selectors to make up a p-address.
*
* Output Parameters
*       None
*
*
* Called By
*       main (initiator/responder)
*
* External Calls
*       ap_set_env()    - an external ACSE/Presentation library function
*                         to set an environment attribute.
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
int set_paddr ( fd, attribute, p, s, t, nsap )
int	fd;		   /* File descriptor*/
int     attribute;         /* Attribute type to be set */
char    *p, *s, *t, *nsap; /* Pointer to p, s, t selectors and NSAP */
{
    int         rc;     /* Return code from ap_set_env */
    ap_paddr_t  p_addr; /* A/P structure to be filled    */
    ap_val_t    val;    /* Value passes to ap_set_env  */
    int afd,i;

    /**************************************************************/
    /* Make up a p_addr using the p, s, t, and NSAPs              */
    /**************************************************************/

    if ( build_paddr ( p, s, t, nsap, &p_addr ) == FAIL )

	return (-1);
    /**************************************************************/
    /* Passes the value to ap_set_env. and frees the inner space. */
    /**************************************************************/


    val.v = (void *) &p_addr;
    rc = ap_set_env ( fd, attribute, val );
    free_paddr ( &p_addr );
 
 
    return ( rc );
 
} /* set_paddr */
 

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        build_paddr()
*
* Description
*       Build a presentation address.
*
* Input Parameters
*       p, s, t, n      - pointer to selector values.
*       p_addr          - composed p-address.
*
* Output Parameters
*       None
*
*
* Called By
*       set_paddr()
*
* External Calls
*       None   
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
build_paddr (p, s, t, n, p_addr)
char            *p, *s, *t, *n;
ap_paddr_t      *p_addr;
{

    char        *mptr;          /* Pointer to memory allocated by this routine */
    int         size;           /* Size of memory needed */
    char        *next;          /* Pointer to available space for next selector */
    char        *real_p, *real_s, *real_t, *real_n;
    int         p_len, s_len, t_len, n_len;
    
    p_addr->n_nsaps = 1;  

    /*******************************************/
    /* Translate addresses in hexdecimal form, */
    /* i.e 0x4852494E42, if necessary.         */
    /*******************************************/
    if (ascii_to_hex(p, &real_p, &p_len)){
	return(-1);
    }
    if (ascii_to_hex(s, &real_s, &s_len)){
	return(-1);
    }
    if ( t != 0 ) {
      if (ascii_to_hex(t, &real_t, &t_len)){
	return(-1);
        }
      } else {
        real_t = 0 ;
        t_len = 0;
        }
    if (ascii_to_hex(n, &real_n, &n_len)){
	return(-1);
    }
 

    /************************************************************/
    /* Allocate space in one big piece for the P-addr structure */
    /* to be filled later.                                      */
    /************************************************************/
    size = ( (sizeof(ap_octet_string_t) * NUM_SELECTORS ) +
              p_len + s_len + t_len + n_len );
    tm_malloc ( size, mptr );
 
    /*****************************************************/
    /* Set the starting memory address for each selector */
    /* and fill in the selector values one by one.       */
    /*****************************************************/

    next = (char *) (mptr + SELECTORS_OFFSET);
    p_addr->p_selector = (ap_octet_string_t *) mptr;
    p_addr->p_selector->length = p_len;
    p_addr->p_selector->data = (unsigned char *)next;
    memcpy ( p_addr->p_selector->data, real_p, p_addr->p_selector->length);
 
    next += p_addr->p_selector->length;
    p_addr->s_selector = (ap_octet_string_t *) (mptr + S_SELECTOR_OFFSET);
    p_addr->s_selector->length = s_len;
    p_addr->s_selector->data   = (unsigned char *)next;
    memcpy ( p_addr->s_selector->data, real_s, p_addr->s_selector->length);
 
    next += p_addr->s_selector->length;
    if ( t_len == 0)  
       { 
       p_addr->t_selector = (ap_octet_string_t *) (0 );
       } 
     else 
       { 
       p_addr->t_selector = (ap_octet_string_t *) (mptr + T_SELECTOR_OFFSET);
       p_addr->t_selector->length = t_len;
       p_addr->t_selector->data   = (unsigned char *)next;
       memcpy ( p_addr->t_selector->data, real_t, p_addr->t_selector->length);
       next += p_addr->t_selector->length;
       }

 
    p_addr->nsaps = (ap_octet_string_t *) (mptr + N_SELECTOR_OFFSET);
    p_addr->nsaps->length      = n_len;
    p_addr->nsaps->data        = (unsigned char *)next;
    memcpy ( p_addr->nsaps->data, real_n, p_addr->nsaps->length);
 
    free(real_p); free(real_s); free(real_t); free(real_n);
 
    return ( SUCCESS );
 
} /* Build paddr */
 

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        ascii_to_hex()
*
* Description
*        Convert ASCII string to hexadecimal string.
*
* Input Parameters
*       ascii_string    - Input ascii string.
*       hex_str_ptr     - Pointer to output hexdecimal string.
*       length          - Length of converted hex string.
*
* Output Parameters
*       None
*
*
* Called By 
*       build_paddr()
*
* External Calls
*       None   
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
ascii_to_hex(ascii_string,hex_str_ptr,length)
char *ascii_string;
unsigned char **hex_str_ptr;
int *length;

{
  int hexdigit;
  int byte;   
  int binequiv;
  unsigned char *binstring;

  if ((ascii_string[0] == '0') && (ascii_string[1] == 'x')) {
    binstring =  *hex_str_ptr = (unsigned char *)malloc(strlen(ascii_string)/2);
    *hex_str_ptr = binstring;
    /****************************************************************************/
    /* Check for even number of hexadecimal digits, if not even, return error   */
    /****************************************************************************/
    if ((strlen(ascii_string) % 2) != 0)
      return(-2);               
 
    for (hexdigit = 2 , byte = 0 ;
       hexdigit < strlen(ascii_string)-1 ;
       hexdigit = hexdigit+2 , byte++) {
 
      /******************************************************************/
      /* For the most significant hexadecimal digit in the current byte */
      /******************************************************************/
      if ((binequiv = HexDigitToBinary(ascii_string[hexdigit])) == -1)
        return(-1);
      else
        binstring[byte] = binequiv*16;
         
      /********************************************************************/
      /* For the least significant hexadecimal digit in the current byte  */
      /********************************************************************/
      if ((binequiv = HexDigitToBinary(ascii_string[hexdigit+1])) == -1)
        return(-1);
      else
        binstring[byte] = binstring[byte]+binequiv;
    }    
    *length = byte;
    binstring[byte] = '\0';     /* delimit the string */
  }
  else {
    binstring = *hex_str_ptr = malloc(strlen(ascii_string)+1);
    strcpy((char *)binstring, ascii_string);
    *length = strlen(ascii_string);
  }
  return(0);
}
 
/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        free_paddr()
*
* Description
*       Free the internal memory of structure ap_paddr_t built by
*       build_paddr() routine.  Memory was allocated in one big
*       piece.
*
* Input Parameters
*       p_addr_ptr      - Pointer to ap_paddr_t structure.
*
* Output Parameters
*       None
*
*
* Called By
*       set_paddr()
*
* External Calls
*       None   
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
free_paddr (p_addr_ptr)
ap_paddr_t      *p_addr_ptr;
{
    char        *starting_ptr;
    int         size = 0;

    if ( p_addr_ptr->p_selector != NULL )
        starting_ptr = (char *)p_addr_ptr->p_selector;
    else if ( p_addr_ptr->p_selector != NULL )
        starting_ptr= (char *)p_addr_ptr->s_selector;
    else if ( p_addr_ptr->t_selector != NULL )
        starting_ptr = (char *)p_addr_ptr->t_selector;
    else if ( p_addr_ptr->nsaps != NULL )
        starting_ptr = (char *)p_addr_ptr->nsaps;
    tm_free ( starting_ptr );
    memset ( p_addr_ptr, 0, sizeof (ap_paddr_t) );
    
} /* Free paddr */

/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        HexDigitToBinary()
*
* Description
*       Convert ASCII hexadecimal digit to binary
*
* Input Parameters
*       hexdigit       - Input hexadecimal digit.
*
* Output Parameters
*       None
*
*
* Called By
*       ascii_to_hex()
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
HexDigitToBinary(hexdigit)
char hexdigit;
{
  int bindigit;
 
  if ((hexdigit >= '0' && hexdigit <= '9'))
    bindigit = hexdigit-'0';
  else
    if ((hexdigit >= 'A' && hexdigit <= 'F'))
      bindigit = (hexdigit-'A')+10;
    else
      if ((hexdigit >= 'a' && hexdigit <= 'f'))
        bindigit = (hexdigit-'a')+10;
      else
        return(-1);             /* Illegal hex digit */
  return(bindigit);
}
/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        Encode_data ()
*
* Description
*       Prepare data buffer to be sent through APLI Interface 
*
* Input Parameters
*       None
*
* Output Parameters
*       None
*
*
* Called By
*       DataApliXfer()
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
Encode_data(buffer , size)
unsigned char *buffer;
int  *size;
{

     int sizeData;
     char line[256];
     int begin,i;
 
     for ( i=0; i< 256; i++) buffer[i]= 0;

 
     printf("Enter your message ( size less than 120 ): ");
     gets(line);
     sizeData= strlen(line);
     printf("\n\n You entered a message of tlength which is %d \n",sizeData);
     buffer[0]= 0x40;
 
     /***************************************/
     /***   Set the size of of the data  ****/
     /***************************************/
     if ( sizeData <= 128 )
        {
        begin = 2;
        buffer[1]=sizeData;
        buffer[2]=0x04;
        *size=sizeData+2;
 
        }
     else
       {
       buffer[1] = 0x82;
       *size= sizeData + 4;
       buffer[2] = (sizeData & 0xff80) >>7;
       buffer[3] = (sizeData & 0x7f) ;
       begin = 4;
       }
        
     sprintf((char *)&buffer[begin],"%s",line);

}



/*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Function Name:        Decode_data ()
*
* Description
*       Prepare data buffer to be sent through APLI Interface 
*
* Input Parameters
*       None
*
* Output Parameters
*       None
*
*
* Called By
*       DataApliXfer()
*
* External Calls
*       None
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/
Decode_data(buffer)
unsigned char *buffer;
{

     int sizeData,size;
     char line[256];
     int begin,i;
 

     if ( buffer[0]!= 0x40 ) 
          printf(" Bad encoded buffer \n");

     if ( buffer[1]!= 0x82 )
        {
        size=buffer[1];
	begin= 2;
        }
     else
        {
        size=buffer[3]+( buffer[2]<< 7);
	begin=4;
        }

     for ( i=0 ; i < 256 ; i++ ) line[i] = 0;

     strncpy(line,(char *)&buffer[begin],size);
     line[size]= 0 ;
     printf(" Size of your message is : %d \n", size);
     printf(" Your message is : 	%s \n\n\n", line);
 
    
}



